@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 import time
 from datetime import datetime
+from typing import Dict
 
 try:
     import ujson as json
@@ -68,6 +69,8 @@ def _make_request(token, method_name, method='get', params=None, files=None):
     :param files: Optional files.
     :return: The result parsed to a JSON dictionary.
     """
+    if not token:
+        raise Exception('Bot token is not defined')
     if API_URL:
         request_url = API_URL.format(token, method_name)
     else:
@@ -166,6 +169,21 @@ def get_me(token):
     return _make_request(token, method_url)
 
 
+def get_my_commands(token):
+    method_url = r'getMyCommands'
+    return _make_request(token, method_url)
+
+
+def log_out(token):
+    method_url = r'logOut'
+    return _make_request(token, method_url)
+
+
+def close(token):
+    method_url = r'close'
+    return _make_request(token, method_url)
+
+
 def get_file(token, file_id):
     method_url = r'getFile'
     return _make_request(token, method_url, params={'file_id': file_id})
@@ -194,7 +212,8 @@ def download_file(token, file_path):
 def send_message(
         token, chat_id, text,
         disable_web_page_preview=None, reply_to_message_id=None, reply_markup=None,
-        parse_mode=None, disable_notification=None, timeout=None):
+        parse_mode=None, disable_notification=None, timeout=None,
+        entities=None, allow_sending_without_reply=None):
     """
     Use this method to send text messages. On success, the sent Message is returned.
     :param token:
@@ -206,6 +225,8 @@ def send_message(
     :param parse_mode:
     :param disable_notification:
     :param timeout:
+    :param entities:
+    :param allow_sending_without_reply:
     :return:
     """
     method_url = r'sendMessage'
@@ -222,6 +243,10 @@ def send_message(
         payload['disable_notification'] = disable_notification
     if timeout:
         payload['connect-timeout'] = timeout
+    if entities:
+        payload['entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -315,6 +340,18 @@ def get_chat_members_count(token, chat_id):
     return _make_request(token, method_url, params=payload)
 
 
+def set_sticker_set_thumb(token, name, user_id, thumb):
+    method_url = r'setStickerSetThumb'
+    payload = {'name': name, 'user_id': user_id}
+    files = {}
+    if thumb:
+        if not isinstance(thumb, str):
+            files['thumb'] = thumb
+        else:
+            payload['thumb'] = thumb
+    return _make_request(token, method_url, params=payload, files=files or None)
+
+
 def set_chat_sticker_set(token, chat_id, sticker_set_name):
     method_url = r'setChatStickerSet'
     payload = {'chat_id': chat_id, 'sticker_set_name': sticker_set_name}
@@ -372,7 +409,7 @@ def copy_message(token, chat_id, from_chat_id, message_id, caption=None, parse_m
 def send_dice(
         token, chat_id,
         emoji=None, disable_notification=None, reply_to_message_id=None,
-        reply_markup=None, timeout=None):
+        reply_markup=None, timeout=None, allow_sending_without_reply=None):
     method_url = r'sendDice'
     payload = {'chat_id': chat_id}
     if emoji:
@@ -385,13 +422,16 @@ def send_dice(
         payload['reply_markup'] = _convert_markup(reply_markup)
     if timeout:
         payload['connect-timeout'] = timeout
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload)
 
 
 def send_photo(
         token, chat_id, photo,
         caption=None, reply_to_message_id=None, reply_markup=None,
-        parse_mode=None, disable_notification=None, timeout=None):
+        parse_mode=None, disable_notification=None, timeout=None,
+        caption_entities=None, allow_sending_without_reply=None):
     method_url = r'sendPhoto'
     payload = {'chat_id': chat_id}
     files = None
@@ -413,13 +453,17 @@ def send_photo(
         payload['disable_notification'] = disable_notification
     if timeout:
         payload['connect-timeout'] = timeout
+    if caption_entities:
+        payload['caption_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(caption_entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_media_group(
         token, chat_id, media,
         disable_notification=None, reply_to_message_id=None,
-        timeout=None):
+        timeout=None, allow_sending_without_reply=None):
     method_url = r'sendMediaGroup'
     media_json, files = convert_input_media_array(media)
     payload = {'chat_id': chat_id, 'media': media_json}
@@ -429,6 +473,8 @@ def send_media_group(
         payload['reply_to_message_id'] = reply_to_message_id
     if timeout:
         payload['connect-timeout'] = timeout
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(
         token, method_url, params=payload,
         method='post' if files else 'get',
@@ -437,14 +483,24 @@ def send_media_group(
 
 def send_location(
         token, chat_id, latitude, longitude,
-        live_period=None, reply_to_message_id=None, reply_markup=None,
-        disable_notification=None, timeout=None):
+        live_period=None, reply_to_message_id=None, 
+        reply_markup=None, disable_notification=None, 
+        timeout=None, horizontal_accuracy=None, heading=None,
+        proximity_alert_radius=None, allow_sending_without_reply=None):
     method_url = r'sendLocation'
     payload = {'chat_id': chat_id, 'latitude': latitude, 'longitude': longitude}
     if live_period:
         payload['live_period'] = live_period
+    if horizontal_accuracy:
+        payload['horizontal_accuracy'] = horizontal_accuracy
+    if heading:
+        payload['heading'] = heading
+    if proximity_alert_radius:
+        payload['proximity_alert_radius'] = proximity_alert_radius
     if reply_to_message_id:
         payload['reply_to_message_id'] = reply_to_message_id
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     if reply_markup:
         payload['reply_markup'] = _convert_markup(reply_markup)
     if disable_notification is not None:
@@ -454,14 +510,22 @@ def send_location(
     return _make_request(token, method_url, params=payload)
 
 
-def edit_message_live_location(token, latitude, longitude, chat_id=None, message_id=None,
-                               inline_message_id=None, reply_markup=None, timeout=None):
+def edit_message_live_location(
+        token, latitude, longitude, chat_id=None, message_id=None,
+        inline_message_id=None, reply_markup=None, timeout=None,
+        horizontal_accuracy=None, heading=None, proximity_alert_radius=None):
     method_url = r'editMessageLiveLocation'
     payload = {'latitude': latitude, 'longitude': longitude}
     if chat_id:
         payload['chat_id'] = chat_id
     if message_id:
         payload['message_id'] = message_id
+    if horizontal_accuracy:
+        payload['horizontal_accuracy'] = horizontal_accuracy
+    if heading:
+        payload['heading'] = heading
+    if proximity_alert_radius:
+        payload['proximity_alert_radius'] = proximity_alert_radius
     if inline_message_id:
         payload['inline_message_id'] = inline_message_id
     if reply_markup:
@@ -492,7 +556,9 @@ def stop_message_live_location(
 def send_venue(
         token, chat_id, latitude, longitude, title, address,
         foursquare_id=None, foursquare_type=None, disable_notification=None,
-        reply_to_message_id=None, reply_markup=None, timeout=None):
+        reply_to_message_id=None, reply_markup=None, timeout=None,
+        allow_sending_without_reply=None, google_place_id=None,
+        google_place_type=None):
     method_url = r'sendVenue'
     payload = {'chat_id': chat_id, 'latitude': latitude, 'longitude': longitude, 'title': title, 'address': address}
     if foursquare_id:
@@ -507,12 +573,19 @@ def send_venue(
         payload['reply_markup'] = _convert_markup(reply_markup)
     if timeout:
         payload['connect-timeout'] = timeout
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
+    if google_place_id:
+        payload['google_place_id'] = google_place_id
+    if google_place_type:
+        payload['google_place_type'] = google_place_type
     return _make_request(token, method_url, params=payload)
 
 
 def send_contact(
         token, chat_id, phone_number, first_name, last_name=None, vcard=None,
-        disable_notification=None, reply_to_message_id=None, reply_markup=None, timeout=None):
+        disable_notification=None, reply_to_message_id=None, reply_markup=None, timeout=None,
+        allow_sending_without_reply=None):
     method_url = r'sendContact'
     payload = {'chat_id': chat_id, 'phone_number': phone_number, 'first_name': first_name}
     if last_name:
@@ -527,6 +600,8 @@ def send_contact(
         payload['reply_markup'] = _convert_markup(reply_markup)
     if timeout:
         payload['connect-timeout'] = timeout
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload)
 
 
@@ -539,7 +614,8 @@ def send_chat_action(token, chat_id, action, timeout=None):
 
 
 def send_video(token, chat_id, data, duration=None, caption=None, reply_to_message_id=None, reply_markup=None,
-               parse_mode=None, supports_streaming=None, disable_notification=None, timeout=None, thumb=None, width=None, height=None):
+               parse_mode=None, supports_streaming=None, disable_notification=None, timeout=None, 
+               thumb=None, width=None, height=None, caption_entities=None, allow_sending_without_reply=None):
     method_url = r'sendVideo'
     payload = {'chat_id': chat_id}
     files = None
@@ -575,11 +651,17 @@ def send_video(token, chat_id, data, duration=None, caption=None, reply_to_messa
         payload['width'] = width
     if height:
         payload['height'] = height
+    if caption_entities:
+        payload['caption_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(caption_entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
-def send_animation(token, chat_id, data, duration=None, caption=None, reply_to_message_id=None, reply_markup=None,
-               parse_mode=None, disable_notification=None, timeout=None, thumb=None):
+def send_animation(
+        token, chat_id, data, duration=None, caption=None, reply_to_message_id=None, reply_markup=None,
+        parse_mode=None, disable_notification=None, timeout=None, thumb=None, caption_entities=None,
+        allow_sending_without_reply=None):
     method_url = r'sendAnimation'
     payload = {'chat_id': chat_id}
     files = None
@@ -609,11 +691,16 @@ def send_animation(token, chat_id, data, duration=None, caption=None, reply_to_m
                 files = {'thumb': thumb}
         else:
             payload['thumb'] = thumb
+    if caption_entities:
+        payload['caption_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(caption_entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_voice(token, chat_id, voice, caption=None, duration=None, reply_to_message_id=None, reply_markup=None,
-               parse_mode=None, disable_notification=None, timeout=None):
+               parse_mode=None, disable_notification=None, timeout=None, caption_entities=None,
+               allow_sending_without_reply=None):
     method_url = r'sendVoice'
     payload = {'chat_id': chat_id}
     files = None
@@ -635,11 +722,15 @@ def send_voice(token, chat_id, voice, caption=None, duration=None, reply_to_mess
         payload['disable_notification'] = disable_notification
     if timeout:
         payload['connect-timeout'] = timeout
+    if caption_entities:
+        payload['caption_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(caption_entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_video_note(token, chat_id, data, duration=None, length=None, reply_to_message_id=None, reply_markup=None,
-                    disable_notification=None, timeout=None, thumb=None):
+                    disable_notification=None, timeout=None, thumb=None, allow_sending_without_reply=None):
     method_url = r'sendVideoNote'
     payload = {'chat_id': chat_id}
     files = None
@@ -669,11 +760,14 @@ def send_video_note(token, chat_id, data, duration=None, length=None, reply_to_m
                 files = {'thumb': thumb}
         else:
             payload['thumb'] = thumb
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_audio(token, chat_id, audio, caption=None, duration=None, performer=None, title=None, reply_to_message_id=None,
-               reply_markup=None, parse_mode=None, disable_notification=None, timeout=None, thumb=None):
+               reply_markup=None, parse_mode=None, disable_notification=None, timeout=None, thumb=None,
+               caption_entities=None, allow_sending_without_reply=None):
     method_url = r'sendAudio'
     payload = {'chat_id': chat_id}
     files = None
@@ -707,11 +801,16 @@ def send_audio(token, chat_id, audio, caption=None, duration=None, performer=Non
                 files = {'thumb': thumb}
         else:
             payload['thumb'] = thumb
+    if caption_entities:
+        payload['caption_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(caption_entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_data(token, chat_id, data, data_type, reply_to_message_id=None, reply_markup=None, parse_mode=None,
-              disable_notification=None, timeout=None, caption=None, thumb=None):
+              disable_notification=None, timeout=None, caption=None, thumb=None, caption_entities=None,
+              allow_sending_without_reply=None, disable_content_type_detection=None):
     method_url = get_method_by_type(data_type)
     payload = {'chat_id': chat_id}
     files = None
@@ -739,6 +838,12 @@ def send_data(token, chat_id, data, data_type, reply_to_message_id=None, reply_m
                 files = {'thumb': thumb}
         else:
             payload['thumb'] = thumb
+    if caption_entities:
+        payload['caption_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(caption_entities))
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
+    if method_url == 'sendDocument' and disable_content_type_detection is not None:
+        payload['disable_content_type_detection'] = disable_content_type_detection
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
@@ -749,13 +854,15 @@ def get_method_by_type(data_type):
         return r'sendSticker'
 
 
-def kick_chat_member(token, chat_id, user_id, until_date=None):
+def kick_chat_member(token, chat_id, user_id, until_date=None, revoke_messages=None):
     method_url = 'kickChatMember'
     payload = {'chat_id': chat_id, 'user_id': user_id}
     if isinstance(until_date, datetime):
         payload['until_date'] = until_date.timestamp()
     else:
         payload['until_date'] = until_date
+    if revoke_messages is not None:
+        payload['revoke_messages'] = revoke_messages
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -804,7 +911,8 @@ def restrict_chat_member(
 def promote_chat_member(
         token, chat_id, user_id, can_change_info=None, can_post_messages=None,
         can_edit_messages=None, can_delete_messages=None, can_invite_users=None,
-        can_restrict_members=None, can_pin_messages=None, can_promote_members=None):
+        can_restrict_members=None, can_pin_messages=None, can_promote_members=None,
+        is_anonymous=None, can_manage_chat=None, can_manage_voice_chats=None):
     method_url = 'promoteChatMember'
     payload = {'chat_id': chat_id, 'user_id': user_id}
     if can_change_info is not None:
@@ -823,6 +931,12 @@ def promote_chat_member(
         payload['can_pin_messages'] = can_pin_messages
     if can_promote_members is not None:
         payload['can_promote_members'] = can_promote_members
+    if is_anonymous is not None:
+        payload['is_anonymous'] = is_anonymous
+    if can_manage_chat is not None:
+        payload['can_manage_chat'] = can_manage_chat
+    if can_manage_voice_chats is not None:
+        payload['can_manage_voice_chats'] = can_manage_voice_chats
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -843,6 +957,50 @@ def set_chat_permissions(token, chat_id, permissions):
     return _make_request(token, method_url, params=payload, method='post')
 
 
+def create_chat_invite_link(token, chat_id, expire_date, member_limit):
+    method_url = 'createChatInviteLink'
+    payload = {
+        'chat_id': chat_id
+    }
+
+    if expire_date is not None:
+        payload['expire_date'] = expire_date
+        if isinstance(payload['expire_date'], datetime):
+            payload['expire_date'] = payload['expire_date'].timestamp()
+
+    if member_limit is not None:
+        payload['member_limit'] = member_limit
+
+    return _make_request(token, method_url, params=payload, method='post')
+
+
+def edit_chat_invite_link(token, chat_id, invite_link, expire_date, member_limit):
+    method_url = 'editChatInviteLink'
+    payload = {
+        'chat_id': chat_id,
+        'invite_link': invite_link
+    }
+
+    if expire_date is not None:
+        payload['expire_date'] = expire_date
+        if isinstance(payload['expire_date'], datetime):
+            payload['expire_date'] = payload['expire_date'].timestamp()
+
+    if member_limit is not None:
+        payload['member_limit'] = member_limit
+
+    return _make_request(token, method_url, params=payload, method='post')
+
+
+def revoke_chat_invite_link(token, chat_id, invite_link):
+    method_url = 'revokeChatInviteLink'
+    payload = {
+        'chat_id': chat_id,
+        'invite_link': invite_link
+    }
+    return _make_request(token, method_url, params=payload, method='post')
+
+
 def export_chat_invite_link(token, chat_id):
     method_url = 'exportChatInviteLink'
     payload = {'chat_id': chat_id}
@@ -989,7 +1147,8 @@ def delete_message(token, chat_id, message_id, timeout=None):
 
 def send_game(
         token, chat_id, game_short_name,
-        disable_notification=None, reply_to_message_id=None, reply_markup=None, timeout=None):
+        disable_notification=None, reply_to_message_id=None, reply_markup=None, timeout=None,
+        allow_sending_without_reply=None):
     method_url = r'sendGame'
     payload = {'chat_id': chat_id, 'game_short_name': game_short_name}
     if disable_notification is not None:
@@ -1000,6 +1159,8 @@ def send_game(
         payload['reply_markup'] = _convert_markup(reply_markup)
     if timeout:
         payload['connect-timeout'] = timeout
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload)
 
 
@@ -1064,7 +1225,7 @@ def send_invoice(
         need_name=None, need_phone_number=None, need_email=None, need_shipping_address=None,
         send_phone_number_to_provider = None, send_email_to_provider = None, is_flexible=None,
         disable_notification=None, reply_to_message_id=None, reply_markup=None, provider_data=None,
-        timeout=None):
+        timeout=None, allow_sending_without_reply=None):
     """
     Use this method to send invoices. On success, the sent Message is returned.
     :param token: Bot's token (you don't need to fill this)
@@ -1092,6 +1253,7 @@ def send_invoice(
     :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total price' button will be shown. If not empty, the first button must be a Pay button
     :param provider_data: A JSON-serialized data about the invoice, which will be shared with the payment provider. A detailed description of required fields should be provided by the payment provider.
     :param timeout:
+    :param allow_sending_without_reply:
     :return:
     """
     method_url = r'sendInvoice'
@@ -1130,6 +1292,8 @@ def send_invoice(
         payload['provider_data'] = provider_data
     if timeout:
         payload['connect-timeout'] = timeout
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     return _make_request(token, method_url, params=payload)
 
 
@@ -1226,15 +1390,17 @@ def upload_sticker_file(token, user_id, png_sticker):
 
 
 def create_new_sticker_set(
-        token, user_id, name, title, png_sticker, emojis,
+        token, user_id, name, title, emojis, png_sticker, tgs_sticker,
         contains_masks=None, mask_position=None):
     method_url = 'createNewStickerSet'
     payload = {'user_id': user_id, 'name': name, 'title': title, 'emojis': emojis}
+    stype = 'png_sticker' if png_sticker else 'tgs_sticker'
+    sticker = png_sticker or tgs_sticker
     files = None
-    if not util.is_string(png_sticker):
-        files = {'png_sticker': png_sticker}
+    if not util.is_string(sticker):
+        files = {stype: sticker}
     else:
-        payload['png_sticker'] = png_sticker
+        payload[stype] = sticker
     if contains_masks is not None:
         payload['contains_masks'] = contains_masks
     if mask_position:
@@ -1242,14 +1408,16 @@ def create_new_sticker_set(
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
-def add_sticker_to_set(token, user_id, name, png_sticker, emojis, mask_position):
+def add_sticker_to_set(token, user_id, name, emojis, png_sticker, tgs_sticker, mask_position):
     method_url = 'addStickerToSet'
     payload = {'user_id': user_id, 'name': name, 'emojis': emojis}
+    stype = 'png_sticker' if png_sticker else 'tgs_sticker'
+    sticker = png_sticker or tgs_sticker
     files = None
-    if not util.is_string(png_sticker):
-        files = {'png_sticker': png_sticker}
+    if not util.is_string(sticker):
+        files = {stype: sticker}
     else:
-        payload['png_sticker'] = png_sticker
+        payload[stype] = sticker
     if mask_position:
         payload['mask_position'] = mask_position.to_json()
     return _make_request(token, method_url, params=payload, files=files, method='post')
@@ -1272,7 +1440,8 @@ def send_poll(
         question, options,
         is_anonymous = None, type = None, allows_multiple_answers = None, correct_option_id = None,
         explanation = None, explanation_parse_mode=None, open_period = None, close_date = None, is_closed = None,
-        disable_notifications=False, reply_to_message_id=None, reply_markup=None, timeout=None):
+        disable_notifications=False, reply_to_message_id=None, allow_sending_without_reply=None,
+        reply_markup=None, timeout=None, explanation_entities=None):
     method_url = r'sendPoll'
     payload = {
         'chat_id': str(chat_id),
@@ -1305,10 +1474,15 @@ def send_poll(
         payload['disable_notification'] = disable_notifications
     if reply_to_message_id is not None:
         payload['reply_to_message_id'] = reply_to_message_id
+    if allow_sending_without_reply is not None:
+        payload['allow_sending_without_reply'] = allow_sending_without_reply
     if reply_markup is not None:
         payload['reply_markup'] = _convert_markup(reply_markup)
     if timeout:
         payload['connect-timeout'] = timeout
+    if explanation_entities:
+        payload['explanation_entities'] = json.dumps(
+            types.MessageEntity.to_list_of_dicts(explanation_entities))
     return _make_request(token, method_url, params=payload)
 
 
